#!/usr/bin/env python
"""
NNTPyPOST - NNTP autoposter.

Copyright 2007 Davide Alberani <da@erlug.linux.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import sys, os, getopt, ConfigParser, nntplib, warnings, time, random, re, \
        socket, tempfile, anydbm
from email.MIMEText import MIMEText
from email import Charset
from email.Utils import formatdate
try: from cStringIO import StringIO
except ImportError: from StringIO import StringIO

# Try to set the locale.
import locale
try: locale.setlocale(locale.LC_ALL, locale.getdefaultlocale()[0] or '')
except (ValueError, locale.Error): pass

PRG_NAME = 'NNTPyPOST'
PRG_VERS = __version__ = '0.7'

HELP = """%s %s
    -f file     use the specified configuration file.
    -v          verbose mode.
    -h          print this help and exit.
""" % (PRG_NAME, PRG_VERS)


def _pmsg(msg, kind, fd=sys.stderr):
    """Internally used to print messages to stdout/stderr."""
    fd.write('%s: %s%s' % (kind, msg, os.linesep))
    fd.flush()

def pwarning(msg):
    """Print a warning message to stderr."""
    _pmsg(msg, kind='W')

def perror(msg, exitCode):
    """Print an error message to stderr and exit."""
    _pmsg(msg, 'E')
    sys.exit(exitCode)

def pinfo(msg):
    """Print an informational message to stdout."""
    if VERBOSE: _pmsg(msg, 'I', fd=sys.stdout)

# - parse command line arguments.

try:
    optlist, args = getopt.getopt(sys.argv[1:], 'hvf:',
                                    ['help', 'verbose', 'file='])
except getopt.error, e:
    optlist = args = []
    pwarning('unable to understand options: "%s"' % str(e))

# Configuration file.
CONF_FILE_NAME = 'nntpypost.conf'
CONF_FILE = []
# /etc/ and /etc/conf.d/
if os.name == 'posix':
    CONF_FILE.append('/etc/' + CONF_FILE_NAME)
    CONF_FILE.append('/etc/conf.d/' + CONF_FILE_NAME)
else:
    # etc subdirectory, for non-unix systems.
    CONF_FILE.append(os.path.join(sys.prefix, 'etc', CONF_FILE_NAME))
# nntpypost.conf file in the user directory.
CONF_FILE += [os.path.join(os.getcwd(), CONF_FILE_NAME),
                os.path.join(os.path.expanduser('~'), CONF_FILE_NAME)]
DOT_CONFIG_FILE = '.' + CONF_FILE_NAME
# .nntpypost.conf in current and user directories.
CONF_FILE += [os.path.join(os.getcwd(), DOT_CONFIG_FILE),
                os.path.join(os.path.expanduser('~'), DOT_CONFIG_FILE)]

VERBOSE = False
TIME = time.localtime()
DATE = time.strftime('%d %h %Y', TIME)
SHORT_DATE = time.strftime('%d/%m/%y', TIME)
ORIGINAL_CHARSETS = Charset.CHARSETS

# Manage command line options and arguments.
for opt in optlist:
    if opt[0] in ('-h', '--help'):
        print HELP
        sys.exit(0)
    elif opt[0] in ('-v', '--verbose'):
        VERBOSE = True
    elif opt[0] in ('-f', '--file'):
        CONF_FILE = opt[1]

if not args:
    print HELP
    perror('specify one or more items to post', 2)


def _normalizePath(path):
    """Normalize a filesystem path."""
    path = os.path.expandvars(path)
    path = os.path.expanduser(path)
    path = os.path.normpath(path)
    return os.path.normcase(path)


def _modifyEncoding(charset, header_enc, body_enc):
    """Modify the encoding of header and body, according to the
    cfg_msg_body_encoding and cfg_msg_headers_encoding options."""
    orig_header_enc, orig_body_enc = ORIGINAL_CHARSETS.get(charset,
                                    (None, None, charset))[0:2]
    enc_values = {'quote-printable': Charset.QP, 'base64': Charset.BASE64,
                    '8bit': None}
    header_enc = enc_values.get(header_enc, orig_header_enc)
    body_enc = enc_values.get(body_enc, orig_body_enc)
    Charset.add_charset(charset, header_enc, body_enc)


class ConfigParserWithCase(ConfigParser.ConfigParser):
    """A case-sensitive parser for configuration files."""
    def optionxform(self, optionstr):
        """Option names are case sensitive."""
        return optionstr

    def _interpolate(self, section, option, rawval, vars):
        """Don't do any interpolation: every substitution is handled later."""
        return rawval


# XXX: provide a default?
conf_file = ConfigParserWithCase()
conf_file.read(CONF_FILE)
if not conf_file.sections():
    print HELP
    perror('unable to find configuration file', 1)

def getOpt(section, option, default='', getter=conf_file.get, warn=True):
    """Return an option from the given section, or default."""
    try:
        return getter(section, option)
    except (ConfigParser.NoOptionError, ConfigParser.NoSectionError), e:
        if warn: warnings.warn(str(e))
    return default

def optIsTrue(section, option, default=False):
    """Return True if the given boolean option is True."""
    return getOpt(section, option, default=default,
                    getter=conf_file.getboolean)

def getHeaders(sect, _excludeOptions=('text', 'text_file')):
    """Return non-default, non-config options."""
    return dict([(o, getOpt(sect, o)) for o in conf_file.options(sect)
            if o not in _excludeOptions and o[:4] not in ('cfg_', 'var_')])


def _specialKeys(key, d):
    """Handle special key values (and/or replace key with values of the
    d dictionary)."""
    _def = '%(' + key + ')s'
    if key.startswith('date_'):
        # Handle a %(date_XYZ)s key.
        req_date = key[5:]
        value = time.strftime(req_date, TIME)
    elif key.startswith('random_'):
        # Handle a %(random_XYZ)s key.
        try:
            req_limit = int(key[7:])
            value = str(random.randrange(req_limit))
        except (ValueError, TypeError):
            value = _def
    else:
        value = d.get(key, _def)
    return value


# The regular expression used to match %(key)s values.
re_subst = re.compile(r'%\((.+?)\)s')

def _substKey(key, d, _seenKeys=None):
    """The d[key] value in the dictionary d is replaced, substituting %(key)s
    placeholders with the value of key(s) in the same d dictionary.
    The whole modified dictionary is returned."""
    # Prevent recursion.
    if _seenKeys is None:
        _seenKeys = []
    if key in _seenKeys:
        return d
    _seenKeys.append(key)
    try:
        d[key] = d[key] % d
    except KeyError:
        # One or more of the %(key)s where not in the d[key].
        for keyNotDone in re_subst.findall(d[key]):
            value = _specialKeys(keyNotDone, d)
            d[key] = d[key].replace('%(' + keyNotDone + ')s',
                                                value)
    # The replacements contained some unreplaced %(key)s, so recur.
    if re_subst.search(d[key]):
        d = _substKey(key, d, _seenKeys)
    return d


def replaceVars(sectOrDict, keysToReturn=[], ignoreAuto=False):
    """Given a string (representing a section name) or a dictionary,
    replace every %(key)s strings in the values of the dictionary;
    returns two dictionary: the first one containing only keys that
    are listed in keysToReturn, the second with every (maybe generated
    by this function) value.
    If ignoreAuto is true, every key with value 'auto' is ignored, to
    prevent propagation of the 'auto' value in other headers."""
    if isinstance(sectOrDict, dict):
        complDict = sectOrDict
    else:
        complDict = dict([(o, getOpt(sectOrDict, o))
                            for o in conf_file.options(sectOrDict)])
    # If not specified, returns everything.
    if not keysToReturn:
        keysToReturn = complDict.keys()
    if not complDict.has_key('date'):
        complDict['date'] = DATE
    if not complDict.has_key('shortDate'):
        complDict['shortDate'] = SHORT_DATE
    _integrate = {}
    # If required, excludes 'auto' values.
    if ignoreAuto:
        _integrate = dict([(k, v) for k, v in complDict.items() if v == 'auto'])
        complDict = dict([(k, v) for k, v in complDict.items() if v != 'auto'])
    # Convert value for every item in the dictionary.
    for key in complDict.keys():
        complDict = _substKey(key, complDict)
    complDict.update(_integrate)
    outDict = {}
    for key in complDict.keys():
        if key not in keysToReturn:
            continue
        outDict[key] = complDict[key]
    return outDict, complDict


def replaceDictInText(text, d):
    """Replace, in text, every occurrence of %(key)s, taking key from the
    d dictionary and handling special cases."""
    def _substFunct(match):
        key = match.group(1)
        return d.get(key, _specialKeys(key, d))
    return re_subst.sub(_substFunct, text)


def _manageSupersedesDB(sect, _write=None):
    """Read and/or set (if _write is not None) a value from/to the database."""
    value = None
    _dbFile = _normalizePath(getOpt(sect, 'cfg_dbfile',
                                    default='.nntpypost.db'))
    try:
        db_fp = anydbm.open(_dbFile, 'c')
        value = db_fp.get(sect)
        if _write is not None:
            db_fp[sect] = _write
        db_fp.close()
    except IOError, e:
        pwarning('unable to read/write file "%s": "%s"' % (_dbFile, e))
    return value


def sendPost(m, sect):
    """Send the post."""
    params = {'host': getOpt(sect, 'cfg_server'), 'readermode': True}
    port = getOpt(sect, 'cfg_port', warn=False)
    if port:
        try: params['port'] = int(port)
        except ValueError: pass
    for key in ('user', 'password'):
        value = getOpt(sect, 'cfg_%s' % key, warn=False)
        if value: params[key] = value
    pinfo('preparing to send post "%s" to %s' % (sect, params['host']))
    post_raw = m.as_string(unixfrom=False)
    # Interactive mode.
    if optIsTrue(sect, 'cfg_interactive'):
        pinfo('entering interactive mode')
        editor = getOpt(sect, 'cfg_editor', warn=False) or \
                os.environ.get('EDITOR') or 'vi'
        temp_file = tempfile.NamedTemporaryFile()
        temp_file.file.write(post_raw)
        temp_file.file.flush()
        os.system('%s "%s"' % (editor, temp_file.name))
        read_file = open(temp_file.name, 'r')
        post_raw = read_file.read()
        read_file.close()
        temp_file.close()
    flike_obj = StringIO(post_raw)
    _messageID = m.get('message-id')
    pinfo('sending post %s' % _messageID or 'unknown Message-ID')
    if not optIsTrue(sect, 'cfg_test'):
        # This is not a test.
        try:
            nntp_server = nntplib.NNTP(**params)
            if optIsTrue(sect, 'cfg_use_ihave'):
                nntp_server.ihave(m.get('message-id'), flike_obj)
            else:
                nntp_server.post(flike_obj)
            # If we have a Message-ID, write it to the database.
            if _messageID is not None:
                _manageSupersedesDB(sect, _write=_messageID)
            nntp_server.quit()
        except (nntplib.NNTPError, socket.error), e:
            perror('Unable to send post "%s": "%s"' % (sect, e), 3)
    else:
        # Just print some information and the whole message.
        print 'I: TEST MODE; sending %s to %s' % (_messageID or
                                                    m.get('subject', '""'),
                                                    m.get('newsgroups'))
        if _messageID is not None:
            print 'I: TEST MODE; assume that the Supersedes database was' \
                    '\nI:            updated with value: %s' % _messageID
        print
        print '=' * 60
        print post_raw
        print '=' * 60
    flike_obj.close()


def modifyHeaders(sect):
    """Modify some headers."""
    pinfo('extracting and normalizing headers')
    cfg_charset = getOpt(sect, 'cfg_conf_charset', default='iso-8859-15',
                        warn=False)
    cfg_msg_charset = getOpt(sect, 'cfg_msg_charset', 'iso-8859-15')
    headersKeys = getHeaders(sect).keys()
    headers, complDict = replaceVars(sect, keysToReturn=headersKeys,
                                    ignoreAuto=True)
    lkeys = [(k.lower(), k) for k in headers.keys()]
    if not filter(lambda x: x[0] == 'message-id', lkeys):
        headers['Message-ID'] = 'auto'
        headersKeys.insert(0, 'Message-ID')
        lkeys.insert(0, ('message-id', 'Message-ID'))
    lkeys.sort()
    for lkey, key in lkeys:
        value = headers[key]
        # Convert the headers to the right charset.
        value = unicode(value, cfg_charset)
        value = value.encode(cfg_msg_charset, 'replace')
        lsvalue = value.strip().lower()
        if lkey == 'group':
            headersKeys.remove('group')
            headersKeys.append('Newsgroups')
            key = 'Newsgroups'
        elif lkey == 'date' and lsvalue == 'auto':
            value = formatdate(localtime=True)
        elif lkey == 'message-id' and lsvalue == 'auto':
            # Forge a brand new Message-ID.
            date_short = time.strftime('%m%d%y-%H%M%S')
            rand_int = random.randint(0, 100000)
            hostname = socket.gethostbyaddr(socket.gethostname())[0]
            value = '<%s.%s.%s@%s>' % (PRG_NAME, date_short,
                                        rand_int, hostname)
        elif lkey == 'supersedes':
            # Get the last Message-ID, so that we can override the old message.
            if value == 'auto':
                value = _manageSupersedesDB(sect)
            if value is None:
                continue
        headers[key] = value
    complDict.update(headers)
    # The second run is done so that we can replace %(key)s whose values
    # where 'auto' (and that now are replaced with something more meaningful).
    headers, complDict = replaceVars(complDict, keysToReturn=headersKeys,
                                        ignoreAuto=False)
    return headers, complDict


def managePost(sect):
    """Prepare the message to send."""
    pinfo('managing post "%s"' % sect)
    if not conf_file.has_section(sect):
        pwarning('unable to read section "%s"' % sect)
        return False
    # Get the body of the message.
    text = getOpt(sect, 'text', warn=False)

    if not text:
        # Try reading the file.
        text_file = getOpt(sect, 'text_file', warn=False)
        if not text_file:
            pwarning('neither text or text_file defined for "%s"' % sect)
            return False
        text_file = _normalizePath(text_file)
        try:
            tfp = file(text_file, 'rt')
            text = tfp.read()
            tfp.close()
        except IOError, e:
            pwarning('unable to read file "%s": "%s"' % (text_file, e))
            return False
    else:
        # We got both "text" and "text_file" options.
        if getOpt(sect, 'text_file', warn=False):
            pwarning('you should not specify both "text" and "text_file"')
            pwarning('"text_file" ignored')
    if not text:
        pwarning('empty message body, skipping "%s"' % sect)
        return False
    # Convert text to the correct charset.
    cfg_charset = getOpt(sect, 'cfg_conf_charset', default='iso-8859-15',
                        warn=False)
    cfg_msg_charset = getOpt(sect, 'cfg_msg_charset', 'iso-8859-15')

    headers, complDict = modifyHeaders(sect)

    # Replace %(key)s in the body.
    text = replaceDictInText(text, complDict)
    # Encode the body in the specified charset.
    text = unicode(text, cfg_charset)
    text = text.encode(cfg_msg_charset, 'replace')
    text = text.replace('\\n', '\n')

    # Force the email package to handle the message's encoding as we like.
    header_enc = getOpt(sect, 'cfg_msg_headers_encoding', warn=False,
                    default='auto')
    body_enc = getOpt(sect, 'cfg_msg_body_encoding', warn=False,
                    default='auto')
    _modifyEncoding(Charset.ALIASES.get(cfg_msg_charset, cfg_msg_charset),
                    header_enc, body_enc)
    # Create the message.
    msg = MIMEText(text, _charset=cfg_msg_charset)
    # Add the headers.
    for key, value in headers.items():
        msg[key.capitalize()] = value
    msg['X-Sender'] = '%s ver:%s' % (PRG_NAME, PRG_VERS)
    sendPost(msg, sect)


if __name__ == '__main__':
    # Run for every command line argument.
    if len(args) == 1 and args[0].lower() == 'all':
        args = conf_file.sections()
    for arg in args:
        managePost(arg)


sys.exit(0)


